package ru.landgrafhomyak.bgtu.networks0.low_level.multithreading

import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose2
import ru.landgrafhomyak.utility.highlevel_try_finally.tryFinallyChain
import ru.landrafhomyak.utility.reference_counter.CloseableReferenceCounter

class CountDownLatch : AutoCloseable {
	private val _refcnt: CloseableReferenceCounter
	private var _counter: ULong
	private val _mutex: Mutex
	private val _condition: Condition

	constructor(initialCounterValue: ULong) {
		this._refcnt = CloseableReferenceCounter("Latch was destroyed")
		this._counter = initialCounterValue
		this._mutex = Mutex()
		safeAutoClose2(onError = this._mutex::close) {
			this._condition = Condition()
		}
	}

	fun await() = this._refcnt.withRef {
		this._mutex.withLock {
			if (this._counter == 0uL)
				return
			this._condition.await(this._mutex)
		}
	}

	fun decrement() = this._refcnt.withRef {
		this._mutex.withLock {
			if (this._counter == 0uL) return
			if (--this._counter == 0uL)
				this._condition.signalAll()
		}
	}

	override fun close() {
		this._refcnt.close("Latch is still in use")
		tryFinallyChain { chain ->
			chain.action { this._condition.close() }
			chain.action { this._mutex.close() }
		}
	}
}